import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
import cartopy.crs as ccrs

def _plot_significance(da, ax, **kwargs):

    da_masked = np.ma.masked_where(da <= 0, da)

    #ax.pcolor(
    #    da.lon, da.lat, da_masked,
    #    #cmap=None,
    #    hatch="oo", alpha=0.,
    #    **kwargs
    #    )
    # use plot to plot markers in the locations where da_masked is not masked
    lon, lat = np.meshgrid(da.lon, da.lat)
    ax.plot(
        lon[~da_masked.mask], lat[~da_masked.mask], marker='.', color='k',
        linestyle='', **kwargs
        )

    
    
def plot_array_to_map(
    da: xr.DataArray, 
    ax=None, title=None, 
    projection=None, 
    mask=None, 
    coastlines_color='black',
    contour=False, 
    contour_color = 'r',
    significance: xr.DataArray = None,
    coastlines_width=2,
    aspect_ratio=1.15,
    fill_nan: float = None,
    pval_marker_size=2,
    pval_marker_color='k',
    **kwargs # used either in contourf or plot
    ):

    # if kwargs add_colorbar is true, create a new variable _add_colorbar and set kwargs['add_colorbar'] = False
    if 'add_colorbar' in kwargs:
        _add_colorbar = kwargs['add_colorbar']
        _cbar_kwargs = kwargs['cbar_kwargs'] if 'cbar_kwargs' in kwargs else {}
        kwargs['add_colorbar'] = False
        kwargs.pop('cbar_kwargs')
    else:
        _add_colorbar = False
        _cbar_kwargs = {}
    
    if mask is not None:
        da = da.where(mask)
        
        if significance is not None:
            significance = significance.where(mask)
            
    if fill_nan is not None:

        if significance is not None:
            significance = significance.where(da.notnull(), 0)
        
        da = da.fillna(fill_nan)

    if projection is None:

        if ax is None:
            ax = plt.axes()
            
        ax.set_aspect(aspect_ratio)

        if contour:
            #im = da.plot.contourf(ax=ax, **kwargs)
            da.plot.contour(ax=ax, colors=contour_color)

        else:
            im = da.plot(ax=ax, **kwargs)
            
        if significance is not None: _plot_significance(significance, ax, markersize=pval_marker_size, markerfacecolor=pval_marker_color)

    else:

        if ax is None:
            ax = plt.axes(projection=projection)
        
        ax.set_aspect(aspect_ratio)  
        
        if contour:
            #im = da.plot.contourf(
            #    ax=ax, transform=ccrs.PlateCarree(), **kwargs
            #    )
            im = da.plot.contour(
                ax=ax, colors=contour_color, transform=ccrs.PlateCarree()
                )
            
            if significance is not None: _plot_significance(significance, ax, transform=ccrs.PlateCarree(), markersize=pval_marker_size, markerfacecolor=pval_marker_color)

        else:
            im = da.plot(
                ax=ax,transform=ccrs.PlateCarree(), **kwargs
                )

            if significance is not None: _plot_significance(significance, ax, transform=ccrs.PlateCarree(), markersize=pval_marker_size, markerfacecolor=pval_marker_color)


        ax.coastlines(color=coastlines_color, linewidth=coastlines_width)
        
    if title is not None:
        ax.set_title(title, fontsize=15, fontweight="bold")
        
    if _add_colorbar:

        cb = plt.colorbar(im, ax=ax, **_cbar_kwargs)
        cb.ax.tick_params(labelsize=30)
        # add kwargs to the colorbar

    return ax, im
