import numpy as np
import time
import math

from statsmodels.tsa.api import ExponentialSmoothing, SimpleExpSmoothing, Holt


def exp_smoothing_ts(
    ts,
    alpha=0.1,
    na_rm=True):
    
    """ Function that performs an exponential smoothing on a time series.
        input: ts: time series (xarray object)
               alpha: smoothing factor
               na_rm: if True, removes nan values from the time series during fitting
        output: ts_smooth: smoothed time series (xarray object)
    """
    ts = np.array(ts)
    if np.isnan(ts).all():
        return np.full(ts.shape, np.nan)
    elif not np.isnan(ts).any() or na_rm:
        ts_smooth = SimpleExpSmoothing(ts).fit(smoothing_level=alpha, optimized=False).fittedvalues
    else:
        exit("Error: missing values in the time series and na_rm=False")
    
    return np.array(ts_smooth)

import numpy as np
from statsmodels.tsa.holtwinters import SimpleExpSmoothing

def exp_smoothing_ts_retro(
    ts,
    na_rm=True,
    alpha=0.1,
    include_actual=False,
    min_points=3
):
    """
    Perform retrospective exponential smoothing with point-matching splits.
    
    Parameters:
    -----------
    ts : numpy.array
        Input time series
    na_rm : bool, optional (default=True)
        Remove NaN values during fitting
    alpha : float, optional (default=0.1)
        Smoothing factor
    include_actual : bool, optional (default=False)
        Include actual point in smoothing window
    min_points : int, optional (default=3)
        Minimum number of points for smoothing
    
    Returns:
    --------
    numpy.array
        Smoothed time series
    """
    # Convert to numpy array and handle all NaN case
    ts = np.array(ts)
    if np.isnan(ts).all():
        return np.full(ts.shape, np.nan)
    
    # Handle NaN values
    if not np.isnan(ts).any() or na_rm:
        # Prepare output array
        ts_loess = np.full(ts.shape, np.nan)
        
        # Create index and valid mask
        x = np.arange(len(ts))
        valid_mask = ~np.isnan(ts)
        x_valid = x[valid_mask]
        ts_valid = ts[valid_mask]
        
        # Iterate through points for splitting
        for i in range(min_points, len(x_valid)):
            # Determine smoothing window
            if include_actual:
                smoothing_window = ts_valid[:i+1]
            else:
                smoothing_window = ts_valid[:i]
            
            # Skip if not enough points
            if len(smoothing_window) < min_points:
                continue
            
            # Perform exponential smoothing
            _fit = SimpleExpSmoothing(smoothing_window).fit(
                smoothing_level=alpha, 
                optimized=False
            )
                
            # Store predicted the actual point
            if include_actual:
                ts_loess[x_valid[i]] = _fit.fittedvalues[-1]
            else:
                ts_loess[x_valid[i]] = _fit.predict(i-1)
                
        # If not removing NaNs, align the output
        if not na_rm:
            ts_loess_aligned = np.full(ts.shape, np.nan)
            ts_loess_aligned[valid_mask] = ts_loess[valid_mask]
            return ts_loess_aligned
        
        return ts_loess
    
    # Raise error if NaNs present and na_rm is False
    raise ValueError("Missing values in time series and na_rm=False")

def loess_ts(
    ts,
    na_rm=True, 
    window=40, 
    degree=1):
    
    """ Function that performs a loess smoothing on a time series.
        The time window is set to 30 points per month.
        input: ts: time series (xarray object)
                na_rm: if True, removes nan values from the time series during fitting
                window: number of points to use for the loess smoothing 
        output: ts_loess: smoothed time series (xarray object)
    """
    ts = np.array(ts)
    if np.isnan(ts).all():
        return np.full(ts.shape, np.nan)
    elif not np.isnan(ts).any() or na_rm:
        # drop nan values
        x = range(len(ts))
        loess = Loess(range(len(ts[~np.isnan(ts)])), 
                            ts[~np.isnan(ts)], 
                            degree=degree)
        ts_loess = []
        for xx in x:
            ts_loess.append(loess.estimate(xx, window=window, degree=degree))
    else:
        exit("Error: missing values in the time series and na_rm=False")
    
            
    return np.array(ts_loess)

def tricubic(x):
    y = np.zeros_like(x)
    idx = (x >= -1) & (x <= 1)
    y[idx] = np.power(1.0 - np.power(np.abs(x[idx]), 3), 3)
    return y

def loess_ts_retro(
    ts,
    na_rm=True, 
    window=40, 
    degree=1,
    min_points=3,
    include_actual=False):
    
    """ Function that performs a loess smoothing on a time series using only previous values.
        input: ts: time series (numpy array)
               na_rm: if True, removes nan values from the time series during fitting
               window: maximum number of previous points to use for the loess smoothing 
               degree: degree of the local polynomial used in LOESS
               min_points: minimum number of points required to perform LOESS (default: degree + 2)
        output: ts_loess: smoothed time series (numpy array)
    """
    ts = np.array(ts)
    if np.isnan(ts).all():
        return np.full(ts.shape, np.nan)
    
    if not np.isnan(ts).any() or na_rm:
        x = np.arange(len(ts))
        valid_mask = ~np.isnan(ts)
        x_valid = x[valid_mask]
        ts_valid = ts[valid_mask]
        
        ts_loess = np.full(ts.shape, np.nan)
        min_points = max(min_points, degree + 2)
        
        for i in range(len(x_valid)):
            if i + 1 < min_points:
                continue
            
            if include_actual:
                start_idx = max(0, i - window + 1)
            else:
                start_idx = max(0, i - window)
            
            try:
                loess = Loess(x_valid[start_idx:i+1], 
                              ts_valid[start_idx:i+1], 
                              degree=degree)
                
                ts_loess[x_valid[i]] = loess.estimate(x_valid[i], 
                                                      window=min(window, i+1), 
                                                      degree=degree)
            except:
                # Fallback to simpler method (e.g., moving average) when SVD fails
                ts_loess[x_valid[i]] = np.mean(ts_valid[start_idx:i+1])
        
        if not na_rm:
            ts_loess_aligned = np.full(ts.shape, np.nan)
            ts_loess_aligned[valid_mask] = ts_loess[valid_mask]
            return ts_loess_aligned
        
        return ts_loess
    else:
        raise ValueError("Error: missing values in the time series and na_rm=False")


class Loess(object):

    @staticmethod
    def normalize_array(array):
        min_val = np.min(array)
        max_val = np.max(array)
        return (array - min_val) / (max_val - min_val), min_val, max_val

    def __init__(self, xx, yy, degree=1):
        self.n_xx, self.min_xx, self.max_xx = self.normalize_array(xx)
        self.n_yy, self.min_yy, self.max_yy = self.normalize_array(yy)
        self.degree = degree

    @staticmethod
    def get_min_range(distances, window):
        min_idx = np.argmin(distances)
        n = len(distances)
        if min_idx == 0:
            return np.arange(0, window)
        if min_idx == n-1:
            return np.arange(n - window, n)

        min_range = [min_idx]
        while len(min_range) < window:
            i0 = min_range[0]
            i1 = min_range[-1]
            if i0 == 0:
                min_range.append(i1 + 1)
            elif i1 == n-1:
                min_range.insert(0, i0 - 1)
            elif distances[i0-1] < distances[i1+1]:
                min_range.insert(0, i0 - 1)
            else:
                min_range.append(i1 + 1)
        return np.array(min_range)

    @staticmethod
    def get_weights(distances, min_range):
        max_distance = np.max(distances[min_range])
        weights = tricubic(distances[min_range] / max_distance)
        return weights

    def normalize_x(self, value):
        return (value - self.min_xx) / (self.max_xx - self.min_xx)

    def denormalize_y(self, value):
        return value * (self.max_yy - self.min_yy) + self.min_yy

    def estimate(self, x, window, use_matrix=False, degree=1):
        n_x = self.normalize_x(x)
        distances = np.abs(self.n_xx - n_x)
        min_range = self.get_min_range(distances, window)
        weights = self.get_weights(distances, min_range)

        if use_matrix or degree > 1:
            wm = np.multiply(np.eye(window), weights)
            xm = np.ones((window, degree + 1))

            xp = np.array([[math.pow(n_x, p)] for p in range(degree + 1)])
            for i in range(1, degree + 1):
                xm[:, i] = np.power(self.n_xx[min_range], i)

            ym = self.n_yy[min_range]
            xmt_wm = np.transpose(xm) @ wm
            beta = np.linalg.pinv(xmt_wm @ xm) @ xmt_wm @ ym
            y = (beta @ xp)[0]
        else:
            xx = self.n_xx[min_range]
            yy = self.n_yy[min_range]
            sum_weight = np.sum(weights)
            sum_weight_x = np.dot(xx, weights)
            sum_weight_y = np.dot(yy, weights)
            sum_weight_x2 = np.dot(np.multiply(xx, xx), weights)
            sum_weight_xy = np.dot(np.multiply(xx, yy), weights)

            mean_x = sum_weight_x / sum_weight
            mean_y = sum_weight_y / sum_weight

            b = (sum_weight_xy - mean_x * mean_y * sum_weight) / \
                (sum_weight_x2 - mean_x * mean_x * sum_weight)
            a = mean_y - b * mean_x
            y = a + b * n_x
        return self.denormalize_y(y)


def main():
    xx = np.array([0.5578196, 2.0217271, 2.5773252, 3.4140288, 4.3014084,
                   4.7448394, 5.1073781, 6.5411662, 6.7216176, 7.2600583,
                   8.1335874, 9.1224379, 11.9296663, 12.3797674, 13.2728619,
                   14.2767453, 15.3731026, 15.6476637, 18.5605355, 18.5866354,
                   18.7572812])
    yy = np.array([18.63654, 103.49646, 150.35391, 190.51031, 208.70115,
                   213.71135, 228.49353, 233.55387, 234.55054, 223.89225,
                   227.68339, 223.91982, 168.01999, 164.95750, 152.61107,
                   160.78742, 168.55567, 152.42658, 221.70702, 222.69040,
                   243.18828])

    loess = Loess(xx, yy)

    for x in xx:
        y = loess.estimate(x, window=7, use_matrix=False, degree=1)
        print(x, y)


if __name__ == "__main__":
    start = time.time()

    main()

    end = time. time()
    print(end - start)
